from functools import reduce
from typing import Union,Callable,Optional,_GenericAlias,_VariadicGenericAlias

def type_check(func):
	def wrapper(*args,**kwargs):
		args_list = [(args[i+1],v,t) if i+1 < len(args) else (kwargs.get(v),v,t) for i,(v,t) in enumerate(func.__annotations__.items())]
		for a,v,t in args_list:
			if (isinstance(t,_VariadicGenericAlias) and not isinstance(a,t)) or (isinstance(t,_GenericAlias) and t.__args__ and not isinstance(a,t.__args__)) or (not isinstance(t,(_GenericAlias,_VariadicGenericAlias)) and not isinstance(a,t)):
				raise TypeError(f"{v} must be {(not isinstance(t,_GenericAlias) and t) or (t.__args__ and ' or '.join(x.__name__ for x in t.__args__)) or 'function'}")
		return func(*args,**kwargs)
	return wrapper

class State:

	def __init__(self,name,msg,type=int,parse_mode=None,back=False,webpage_preview=False):
		self.name = name
		self.data_type = type
		self.msg = msg
		self.mode = parse_mode
		self.callback = None
		self.text = None
		self.action = None
		self.build = None
		self.back = back
		self.webpage_preview = webpage_preview
		self.routes = None
		self.custom = None
		self.list = None
		self.list_all = None
		self.list_start = None
	
	def __str__(self):
		return f'State <{self.name}>'

	@type_check
	def add_keyboard(self,keyboard:Union[list,dict],size:Optional[tuple]=None,max_row:Optional[int]=3):
		'''Add inline keyboard for the state'''
		if isinstance(keyboard,list): keyboard = dict(enumerate(keyboard))
		if size and sum(size) != len(keyboard): raise ValueError(f'Keyboard length ({len(keyboard)}) must be the same size as the sum of row size ({sum(size)}).')
		if size: size = reduce(lambda x,y:x+y,map(lambda x:[x] if x < 9 else [8]*(x//8)+([x%8],[])[not x%8],size))
		elem_n = len(keyboard)
		if not size: size = [max_row for _ in range(elem_n//max_row)] + ([r,] if (r := elem_n % max_row) else [])
		self.callback = (keyboard,tuple(size))

	@type_check
	def add_text(self,regex:Optional[str]=None,error:Optional[str]=None):
		'''Add handler for text input'''
		if not regex: regex = r'^.*$' 
		self.text = (regex,error)

	@type_check
	def add_action(self,function:Callable):
		'''Add action to do with this command'''
		self.action = function

	@type_check
	def add_dynamic_keyboard(self,function:Callable,max_row:Optional[int]=3):
		'''Add action to build a keyboard dynamically'''
		self.build = function
		self.max_row = max_row

	@type_check
	def add_custom_keyboard(self,function:Callable):
		'''Add function to build custom keyboard | must return a list of InlineKeyboardButton'''
		self.custom = function

	@type_check
	def add_dynamic_routes(self,function:Callable):
		'''Add function to create dynamic routes | must return 3 value (routes,default,back)'''
		self.routes = function

	@type_check
	def add_dynamic_list(self,function:Callable,start:Optional[int]=0,left_button:Optional[str]='<',right_button:Optional[str]='>',all_pages:Optional[bool]=False):
		'''Add function to create a list with page'''
		self.list = function
		self.list_buttons = [left_button,right_button]
		self.list_all = all_pages
		self.list_start = start



from autoconv.conversation import Conversation
from telegram import InlineKeyboardButton, InlineKeyboardMarkup
from telegram.ext import ConversationHandler
from re import match
from functools import reduce

class AutoConvHandler:

	def __init__(self,conversation,telegram_state_name,back_button='Back'):
		self.conversation = conversation
		self.NEXT = telegram_state_name
		self.back_button = back_button
		self.prev_state = None
		self.curr_state = conversation.start
		self.update, self.context = None, None
		self._bkup_state_routes,self._bkup_state_keyboard = None,None

	def _build_keyboard(self,state):
		'''Build Keyboard for callback state'''
		cmd_list = [[InlineKeyboardButton(text=key_param[0][k],callback_data=k) for k in list(key_param[0].keys())[su:su+si]] for si,su in zip(key_param[1],[sum(key_param[1][:i]) for i in range(len(key_param[1]))])] if (key_param := state.callback) else [[]]
		if state.back: cmd_list += [[InlineKeyboardButton(text=self.back_button,callback_data='BACK')]]
		return (state.custom and state.custom(self.update,self.context)) or cmd_list

	def _next_state(self,state,value):
		'''Follow state ruote'''
		if value not in self.conversation.routes.get(state.name) and -1 not in self.conversation.routes.get(state.name): raise ValueError(f'Deafult route not found and value {value} doesn\'t exist as route of {state}.') 
		return self.conversation.routes.get(state.name).get(value) if value in self.conversation.routes.get(state.name) else self.conversation.routes.get(state.name).get(-1)

	def _change_state(self,telegram_id,data):
		'''Set variables for next state'''
		data_context = self.context.user_data.get(telegram_id)
		state = self.conversation.get_state(self.context.user_data.get(telegram_id).get('state'))
		if state.list:
			list_idx = data if state.list_all or data < 2 else data - len(data_context.get('list'))+2 - (0,1)[data_context.get('list_i') in (0,len(data_context.get('list'))-1)]
			if state.list_all: list_idx = data if data_context.get('list_i') > data else data-1
			value = reduce(lambda x,y: x+y,self.list_keyboard)[list_idx].text
		else: value = d if (c := state.callback) and (d := c[0].get(data)) else data
		if state != self.conversation.end: self.context.user_data.get(telegram_id).get('data').update({state.name:value})
		new_state = self._next_state(state,data)
		self.prev_state,self.curr_state = self.curr_state,new_state
		data_context.update({'prev_state':self.prev_state.name,'state':new_state.name})
		if new_state.text: data_context.update({'error':False})
		elif data_context.get('error') != None: data_context.pop('error')
		return new_state

	def _wrong_text(self):
		'''Handler for wrong regex in text state'''
		telegram_id = self.update.message.chat.id
		self.update.message.delete()
		state = self.conversation.get_state(self.context.user_data.get(telegram_id).get('state'))
		if state.text and not self.context.user_data.get(telegram_id).get('error'):
			keyboard = self._build_keyboard(state)
			self.context.user_data.get(telegram_id).update({'error':True})
			self.context.user_data.get(telegram_id).get('bot-msg').edit_text(f"{state.msg}\n\n{state.text[1]}",reply_markup=InlineKeyboardMarkup(keyboard),parse_mode=state.mode)
		return self.NEXT

	def _going_back(self):
		'''Handler for back button'''
		query = self.update.callback_query
		telegram_id = self.update.callback_query.message.chat.id
		new_state = self._change_state(telegram_id,'BACK')
		keyboard = self._build_keyboard(new_state)
		query.edit_message_text(f"{new_state}",reply_markup=keyboard,parse_mode=new_state.mode)
		return self.NEXT

	def _update_dynamic_list(self,state):
		'''Update i and routes backup for dynamic list'''
		data = self.context.user_data.get(self.update.effective_chat.id)
		if (state_l := data.get('list')):
			i = int(data.get('list_i'))
			new_i = state_l.index(data.get('data').get(state.name)) if state.list_all else (i-1,i+1)[data.get('data').get(state.name) == state.list_buttons[1] or not i]
			data.update({'list_i':new_i})
		else:
			state_l = state.list(self.update,self.context)
			data.update({'list':state_l,'list_i':state.list_start})
		if self._bkup_state_routes:
			self.conversation.add_routes(self.curr_state,self._bkup_state_routes)
			self._bkup_state_routes = None

	def _build_dynamic_routes(self,state):
		'''Build dynamic routes for current state'''
		ro,de,ba = state.routes(self.update,self.context)
		self.conversation.add_routes(state,ro,de,ba)

	def _build_dynamic_keyboard(self,state):
		'''Build dynamic keyboard for current state'''
		keyboard = state.build(self.update,self.context)
		keyboard,size = keyboard if isinstance(keyboard,tuple) else (keyboard,None)
		state.add_keyboard(keyboard,size,max_row=state.max_row)

	def _build_dynamic_list(self,state,keyboard):
		'''Build dynamic list for current state'''
		self._bkup_state_routes = self.conversation.routes.get(state.name)
		data = self.context.user_data.get(self.update.effective_chat.id)
		state_l = data.get('list')
		basic_routes = {k+len(state_l):v for k,v in self.conversation.routes.get(state.name).items()}
		for kl in keyboard:
			for button in kl: 
				if button.callback_data: button.callback_data += len(state_l)
		list_buttons = [InlineKeyboardButton(b,callback_data=i) for i,b in enumerate(state_l if state.list_all else state.list_buttons)]
		keyboard = [list_buttons]+keyboard
		self.list_keyboard = keyboard
		self.conversation.add_routes(state,basic_routes,default=state)
		if not state.list_all and (i := data.get('list_i')) in (0,len(state_l)-1): keyboard[0].pop((1,0)[not i])
		if state.list_all: keyboard[data.get('list_i')//8].pop(data.get('list_i'))
		return keyboard

	def _build_dynamic_stuff(self,state): 
		'''Compute dynamic stuff: action > routes > keyboard > list'''
		data = self.context.user_data.get(self.update.effective_chat.id)
		if self.prev_state != self.curr_state and data.get('list'): data.pop('list'),data.pop('list_i')
		if state.list: self._update_dynamic_list(state)
		action_str = state.action(self.update,self.context) if state.action else None
		if state.routes: self._build_dynamic_routes(state)
		if state.build: self._build_dynamic_keyboard(state)
		keyboard = self._build_keyboard(state)
		if state.list: keyboard = self._build_dynamic_list(state,keyboard)
		reply_msg = state.msg if action_str == None else state.msg.replace('@@@',action_str)
		return InlineKeyboardMarkup(keyboard),reply_msg

	def restart(self):
		'''Restart handler to initial configuration'''
		if self.update:
			telegram_id = self.update.effective_chat.id
			if (c := self.context.user_data.get(telegram_id)):
				if (m := c.get('bot-msg')): m.delete()
				self.context.user_data.pop(telegram_id)
			self.prev_state = None
			self.curr_state = self.conversation.start
		return self

	def manage_conversation(self,update,context,delete_first=True):
		'''Master function for converastion'''
		self.update = update
		self.context = context
		telegram_id = self.update.effective_chat.id
		# start
		if not self.context.user_data.get(telegram_id):
			if delete_first: update.message.delete()
			state = self.conversation.start
			self.context.user_data.update({telegram_id:{'prev_state':None,'state':state.name,'error':False,'data':{}}})
			keyboard,reply_msg = self._build_dynamic_stuff(state)
			msg = self.update.message.reply_text(f'{reply_msg}',reply_markup=keyboard,parse_mode=state.mode)
			self.context.user_data.get(telegram_id).update({'bot-msg':msg})
			return self.NEXT
		# get data
		state = self.conversation.get_state(self.context.user_data.get(telegram_id).get('state'))
		if self.update.callback_query:
			data = self.update.callback_query.data
			to_reply = self.update.callback_query.edit_message_text
		else:
			data = self.update.message.text
			if not state.text: self.update.message.delete(); return self.NEXT
			if not match(state.text[0],data): return self._wrong_text()
			to_reply = self.context.user_data.get(telegram_id).get('bot-msg').edit_text
			self.update.message.delete()
		# next stage
		if not self.conversation.routes.get(state.name) and state.routes: self._build_dynamic_routes(state)
		typed_data = state.data_type(data) if data != 'BACK' else 'BACK'
		state = self._change_state(telegram_id,typed_data)
		keyboard,reply_msg = self._build_dynamic_stuff(state)
		to_reply(f'{reply_msg}',reply_markup=keyboard,parse_mode=state.mode,disable_web_page_preview=not state.webpage_preview)
		if state == self.conversation.end: context.user_data.update({telegram_id:None}); return ConversationHandler.END
		return self.NEXT


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="telegram-autoconv",
    version="0.0.23",
    author="Moris Doratiotto",
    author_email="moris.doratiotto@gmail.com",
    description="A python module to auto build conversation for a Telegram bot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mortafix/AutoConv-Telegram-Python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    ],
    python_requires='>=3.8',
    keywords=['conversation', 'generator', 'telegram', 'bot'],
)